#include "hbnet.h"
#include <stdlib.h>
#include <vector>
#include <set>
#include <list>
#include <algorithm>
#include <errno.h>
#include <stdio.h>

//#define TEST_DROPPED_PACKETS

/*
 *  OUTLINE OF BASIC HBNET SERVER 
 *
 *  hbNetOpen();
 *  while(1)
 *      //  read
 *      hbNetUpdate();
 *      if(data)
 *          process(); 
 *      //  write
 *      hbSendTo();
 *  hbNetClose();
 *
 */

/*
 *  OUTLINE OF BASIC HBNET CLIENT
 *
 *  hbNetOpen();
 *  takeDotAim(); 
 *  while(1)
 *      //  read
 *      hbNetUpdate();
 *      if(data)
 *          process(); 
 *      //  write
 *      hbSend();
 *  hbNetClose();
 *
 */

using namespace std;

//////////////////////////////////////////////////////////////////////////////
HBaddr::HBaddr(void)
{
}

HBaddr::HBaddr(char *dotip, int p)
{
    host = inet_addr(dotip);
    port = p;
}

HBaddr::HBaddr(struct sockaddr_in addr)
{
    host = addr.sin_addr.s_addr;
    port = ntohs(addr.sin_port);
}

bool HBaddr::operator==(HBaddr& x){return host == x.host && port == x.port;}

//////////////////////////////////////////////////////////////////////////
FILE *log;
int  (*lostRemote)(HBaddr); 
bool activated = false;
//  TODO associate this var with a more logical class
socklen_t remoteAddLen = sizeof(struct sockaddr_in);

#define WATCH_INTERVAL 2
#define TRIPS_WATCHED 10

const int MAX_REMOTES = 500;
   
//////////////////////////////////////////////////////////////
class SavedMsg
{
    public:
    SavedMsg(void *m, int l, int s)
    {
        memcpy(msg, m, l);
        msg[l] = 0;
        len = l;
        seqno = s;
    }
    SavedMsg(int s)
    {
        seqno = s;
    }
    char msg[APP_MSG_SIZE + 1]; // TODO need a copy con?
    int  len;
    int  seqno;

    bool operator<(const SavedMsg& sm) { return seqno < sm.seqno; }
};

struct compSavedMsg{
    bool operator()(const SavedMsg& sm1, const SavedMsg& sm2) const {
        return sm1.seqno < sm2.seqno;
    }
};

////////////////////////////////////////////////////////////////////
class WaitForAck: public SavedMsg
{
    public:
    WaitForAck(void *m, int l, int s, timeval t) : SavedMsg(m, l, s)
    {
        timeSent = t;
        sends    = 0;
    }
    WaitForAck(int s) : SavedMsg(s) { }

    bool operator==(const WaitForAck& wfa) { return seqno == wfa.seqno; }

    timeval  timeSent;
    int sends;
};


////////////////////////////////////////////////////////
struct Remote
{
    Remote(HBaddr);
    bool operator==(const Remote& rem);

    HBaddr                          address;
    int                             outboundSeqno; 
    int                             expected;
    list<WaitForAck>                wfa;  
    set<SavedMsg, compSavedMsg>     prm; 
    timeval                         lastReceive;
    timeval                         lastSend;
    long                            tripTime[TRIPS_WATCHED];
    int                             tripndx;
    long                            waitTime;
    int                             watching;
    bool                            valid;
};

Remote::Remote(HBaddr addr)
{
    address         = addr;
    outboundSeqno   = 0;
    expected        = 0;  
    for(int z = 0; z < TRIPS_WATCHED; z++)
        tripTime[z] = 250000l;  // 1/4 second
    waitTime        = 500000l;  //  1/2 second 
                                //  (waitTime is double average of trips)
    tripndx         = 0;
    watching        = -1; 
    valid = true;
}
bool Remote::operator==(const Remote& rem)
{ 
    //return (address == rem.address); //  why is this a compile error?
    HBaddr a = rem.address;
    return address == a;
}

list<Remote> remote;

//  TODO  this looks backwards - should calc app msg size by subtracting header size from total max size, but app msg size should be the only one of the three v in the app-visible header file
const int PACKAGE_SIZE  = 460;
const int HEADER_SIZE   = PACKAGE_SIZE - APP_MSG_SIZE;
const int APP_MSG       = 0;
const int HB_ACK        = 1;
const int HEARTBEAT     = 2;
const int APP_MSG_CLESS = 3;
const int MAX_PREMATURE = 100;
const int MAX_UNACKED   = 100;

int sockfd;
struct sockaddr_in localAddr;
struct sockaddr_in targetAddr;


list<Remote>::iterator remoteHandle(HBaddr, int);
int addPremature(list<Remote>::iterator, int, void*, int);


//////////////////////////////////////////////////////////////////////////////
int hbNetOpen(int port, int ((*stranded)(HBaddr)), char *lg)
{
    int b;
    int x, y;

    activated = true;

    srand(time(NULL));

    if(lg == NULL){
        log = NULL;
    }
    else{
        log = fopen(lg, "w");
    }

#ifdef HBNET_WINDOWS
    WSADATA wsaData;

    if(WSAStartup(MAKEWORD(2, 0), &wsaData) != 0){
        puts("Error: could not init winsock");
        return -1;
    }
#endif

    lostRemote = stranded;

    //  Open a socket and binds it to a port
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sockfd  != -1) {
        //  Bind socket to a port    
        memset(&localAddr, 0, sizeof(localAddr));
        localAddr.sin_family       = AF_INET;    /* host byte order */
        localAddr.sin_addr.s_addr  = INADDR_ANY; /* auto-fill with my IP */
        localAddr.sin_port         = htons(port);/* short, network byte order */
        b = bind(sockfd, (struct sockaddr*)(&localAddr), sizeof(struct sockaddr));
        if(b == -1) {
            sockfd = -1;
        }
    }


    return sockfd;
}

//////////////////////////////////////////////////////////////////////////////
void takeDotAim(char* dotip, int port)
{
    HBaddr addr(dotip, port);
    takeAim(addr);
}


//////////////////////////////////////////////////////////////////////////////
void takeAim(HBaddr addr)
{
//  This function is always succssful, but the send that called this
//  might fail if send is connected and target is new and remote[] is full
    memset(&targetAddr, 0, sizeof(targetAddr));
    targetAddr.sin_family      = AF_INET;      /* host byte order */
    targetAddr.sin_addr.s_addr = addr.host;
    targetAddr.sin_port        = htons(addr.port); /* short, ntwrk byte order */
}

//////////////////////////////////////////////////////////////////////////////
int hbSendTo(HBaddr addr, char* msg, int len, int connected)
{
    takeAim(addr);
    return hbSend(msg, len, connected);    
}

//////////////////////////////////////////////////////////////////////////////
int hbSend(char* msg, int len, int connected)
{
    int              bytesSent;
    char             package[PACKAGE_SIZE];    
    int             *type       = (int*)package;
    int             *seqno      = ((int*)package + 1);

    struct sockaddr *target     = (struct sockaddr*)(&targetAddr);
    HBaddr asHBaddr(targetAddr);

    list<Remote>::iterator rem_itr;
    int              w;
    char             str[33];
    timeval          now;

    if(strlen(msg) > APP_MSG_SIZE)
        return 0;
    if(!activated)
        return 0;

    log && fprintf(log, "\nSending to: %s \n", asHBaddr.toString(str));

    gettimeofday(&now, NULL);

    if(connected == 0){
        rem_itr = remoteHandle(asHBaddr, APP_MSG_CLESS);
        if(rem_itr != remote.end())
            rem_itr->lastSend = now;
        //  Build package
        *type = APP_MSG_CLESS;
        memcpy(package + HEADER_SIZE, msg, len);
        //  Send package
        bytesSent = sendto(sockfd, package, HEADER_SIZE + len, 0, 
            target, remoteAddLen);
        //  Post-send activities
        int bs = bytesSent - HEADER_SIZE;
        log && fprintf(log, "    Unconnected bytes sent = %d\n", bs);
        return bs;
    }

    rem_itr = remoteHandle(asHBaddr, APP_MSG);
    if(rem_itr == remote.end()){
        log && fprintf(log, "ERROR: hbSend cannot send to %sd\n", 
                             asHBaddr.toString(str) );
        return 0;
    }
    if(rem_itr->wfa.size() == MAX_UNACKED){
       return 0;    
       log && fputs("SEND FAILED: max number of unackeds has been reached",log);
    }

    //  Build package
    *type = APP_MSG;
    *seqno = rem_itr->outboundSeqno;
    memcpy(package + HEADER_SIZE, msg, len);

    //  Send package
#ifdef TEST_DROPPED_PACKETS
    int rnd;
    rnd = rand() % 4;
    if (rnd != 0){
        bytesSent = sendto(sockfd, package, HEADER_SIZE + len, 0, 
            target, remoteAddLen);
    }
    else{
       log && fprintf(log, "    TEST: emulating loss of sent msg %d\n", *seqno);
       bytesSent = len + HEADER_SIZE;
    }
#else
    bytesSent = sendto(sockfd, package, HEADER_SIZE + len, 0, 
        target, remoteAddLen);
#endif

    if(bytesSent == 0){
        log && fprintf(log, "    SEND ERROR: bytesSent = 0\n");
        return 0;
    }

    //  Post-send activities
    //log && fprintf(log, "    Remote %d    Message %d sent\n", t, *seqno);
    //  TODO use the better one above, formatted for new container types, etc
    log && fprintf(log, "    sent message %d\n", *seqno);
    rem_itr->lastSend = now;
    WaitForAck uu(msg, len, *seqno, now);
    rem_itr->wfa.push_back(uu);
    //  time a packet round trip once in a while
    w = *seqno % WATCH_INTERVAL;
    if(w == 0 && rem_itr->watching == -1){
        rem_itr->watching = *seqno;
        log && fprintf(log, "    watching message %d\n", *seqno);
    }
    rem_itr->outboundSeqno++;
    return bytesSent - HEADER_SIZE; /*  This should be equal to len,
        but just in case there's something screwy about the send,
        let's be honest about it and dynamically calc it  */
}

//////////////////////////////////////////////////////////////////////////////
int hbNetUpdate(HBaddr* sender, void* msg)
//  FOUR JOBS:
//  1) recvfrom
//  2) resend unacknowledged messages
//  3) send heartbeats
//  4) check for disconnects
{
    fd_set              rfds;
    struct timeval      tv;
    struct timeval      now;
    int                 readPort;
    int                 bytesRead = 0;
    char                package[PACKAGE_SIZE];
    int                *type = (int*)package;
    int                *seqno = ((int*)package + 1);
    struct sockaddr_in *resendAddress;  
    struct sockaddr    *baseAddress;
    struct HBaddr       source;
    int                 n, y, z, r;
    socklen_t           addrLen;
    char str[33];
    list<Remote>::iterator        rrr;
    list<WaitForAck>::iterator    wfa_itr;
    static HBaddr                 gotDelivery;
    static bool                   checkPremature = false; 

    if(!activated)
        return 0;

/*    if(checkPremature != -1){
        Remote& rmt = remote[checkPremature]; */
    if(checkPremature){  
        log && fprintf(log, "\nChecking premature msgs...\n");
        Remote& rmt = *(find(remote.begin(), remote.end(), gotDelivery));
        if(!rmt.prm.empty()){
            set<SavedMsg>::iterator firstPrem = rmt.prm.begin();
            if(firstPrem->seqno == rmt.expected){
                log && fprintf(log, "    Pulling %d from prem set\n", 
                                         rmt.expected);
                memcpy(msg, firstPrem->msg, firstPrem->len);
                rmt.expected++; 
                //  already ACKed, already removed hbnet header and adj size
                bytesRead = firstPrem->len;
                rmt.prm.erase(firstPrem);
                return bytesRead;
            }
            log && fprintf(log, "\n    No go. firstPrem->seqno = %d\n", 
                                       firstPrem->seqno);
        }
        else{	
            log && fprintf(log,
                "    msg %d is not in the prem set\n", rmt.expected);
        }
        checkPremature = false;
    }
    
    FD_ZERO(&rfds);
    FD_SET(sockfd, &rfds);
    //  TODO are these next 2 lines necessary?
    tv.tv_sec = 0;
    tv.tv_usec = 0;

    gettimeofday(&now, NULL); /* get it over with, because even if there's no
    data read from a connection, still might be needed in processing unackeds */

    //  Check for a package queued in socket
    struct sockaddr_in returnAddr;
    baseAddress = (struct sockaddr*)(&returnAddr);
    readPort = select(sockfd + 1, &rfds, NULL, NULL, &tv);
    if(readPort > 0){
        log && fprintf(log, "\nReceived a packet\n");
        addrLen = sizeof(struct sockaddr);
        bytesRead = recvfrom(sockfd, package, PACKAGE_SIZE, 0,
            baseAddress, &addrLen);
        if(bytesRead != -1){
        //  There is data in the socket 
            source = HBaddr(returnAddr);
            *sender = source;
            //x = remoteIndex(source, *type);
            rrr = remoteHandle(source, *type);
                // adds source to remote list if necessary and if type allows
            if(rrr != remote.end())
                rrr->lastReceive = now; /*  That's all we need from a 
                    HEARTBEAT, so we don't need to check for HEARTBEAT type 
                    at all; the job's just been done */
	    if(*type == APP_MSG_CLESS){
                log && fprintf(log, "    Packet is connectionless message\n");
                package[bytesRead] = 0;
                memcpy(msg, package + HEADER_SIZE, bytesRead - HEADER_SIZE);
                log && fprintf(log,
                    "    message copied for processing by app\n");
            }
            else{
            //  message is of some connected type
                if(rrr == remote.end()) //  source is new, no room in remote[]
                    goto check_unacked;
                Remote &rmt = *rrr;
                if(*type == APP_MSG){
                    package[bytesRead] = 0;
                    log && fprintf(log, 
                 //  TODO put dot ip in following nessage
                 "    Cnctd App Msg from xxxxxx seqno: %d  expected: %d\n", 
                          *seqno, rmt.expected); 
                    if(*seqno <= rmt.expected){
                    //  Message definitely requires an ack
                        *type = HB_ACK;
                        addrLen = sizeof(struct sockaddr_in);
#ifdef TEST_DROPPED_PACKETS
                        int rnd;
                        rnd = rand() % 4;
                        if(rnd != 0){
                            sendto(sockfd, package, HEADER_SIZE, 0,
                                baseAddress, addrLen);
                        }
                        else{
                            log && fprintf(log,
                                "        TEST: emulating lost ack\n");
                        }
#else
                        sendto(sockfd, package, HEADER_SIZE, 0,
                            baseAddress, addrLen);
#endif
                        log && fprintf(log, "        sent ack\n");
                        if(*seqno == rmt.expected){
                        //  App needs to process message
                            memcpy(msg, package + HEADER_SIZE, 
                            bytesRead - HEADER_SIZE);
                            rmt.expected++;
                            checkPremature = true;
                            gotDelivery = source;
                            log && fprintf(log,
                              "        message copied for processing by app\n");
                        }
                        else{  // seqno < expected
                            log && fprintf(log,
                                "        app has already processed message\n");
                            bytesRead = 0;  //  return value
                        }
	    	    }  
                    else{  //  seqno > expected
                        log && fprintf(log,  "        Packet is premature\n");
                        n = addPremature(rrr, *seqno, 
                            package + HEADER_SIZE, bytesRead - HEADER_SIZE);
                        if(n){
                            *type = HB_ACK;
                            addrLen = sizeof(struct sockaddr_in);
                            sendto(sockfd, package, HEADER_SIZE, 0,
                                baseAddress, addrLen);
                            log && fprintf(log, "        sent ack\n");
                        }
                        bytesRead = 0;
	            }	
                    log && fflush(log); 
                }
                else if(*type == HB_ACK){
                //  Remove from remote's Unacked list
                    log && fprintf(log, 
                        "    Ack from %s    msg %d\n", 
                             source.toString(str), *seqno);
                    log && fprintf(log, "    watching = %d\n", rmt.watching);
                    if(*seqno == rmt.watching){
                    //  process the timed round trip
                        wfa_itr = find(rmt.wfa.begin(), rmt.wfa.end(), *seqno);
                        long usec = tvSubtraction(now, wfa_itr->timeSent);
                        log && fprintf(log, "        trip time = %ld\n", usec);
                        rmt.tripTime[rmt.tripndx] = usec;
                        rmt.tripndx++;
                        if(rmt.tripndx == TRIPS_WATCHED)
                            rmt.tripndx = 0;
                        //  average rmt.tripTimes:
                        long total = 0;
                        for(int z = 0; z < TRIPS_WATCHED; z++)
                            total += rmt.tripTime[z];
                        rmt.waitTime = total / TRIPS_WATCHED * 2;
                        rmt.watching = -1;
                        log && fflush(log); 
                    }
                    rmt.wfa.remove(*seqno);
                    log && fflush(log); 
                }  // End of processing HB_ACK
	        else if(*type != HEARTBEAT){ 
                    log && fprintf(log, 
                        "    ERROR: Invalid message type\n");
                }
            }
        }
	else{
            log && fprintf(log, "    Port indicated there was a message, but there were no bytes read from it\n");
        }
    }

check_unacked:
    //  Check for unacknowledged packages which need to be resent
    //  Check for timed-out (lost) connections
    //  Send heartbeats   
    resendAddress = &(targetAddr);
    baseAddress = (struct sockaddr*)resendAddress;
    //  nada but sending now, TODO: make two global vars, one for rcv addrLen
    //  and one for send addrlen
    addrLen = sizeof(struct sockaddr_in);
    long timeWaited;
    for(rrr = remote.begin(); rrr != remote.end(); rrr++){

        if(!rrr->valid) 
            continue;

        if(!rrr->wfa.empty()){
            wfa_itr = rrr->wfa.begin();
            timeWaited = tvSubtraction(now, wfa_itr->timeSent);
            if(timeWaited > rrr->waitTime){
                log && fprintf(log, "Trying a resend\n    waitTime = %ld\n    timeWaited = %ld\n", rrr->waitTime, timeWaited);
                wfa_itr->sends++; 
                if(wfa_itr->sends == 11){
                    (*lostRemote)(rrr->address);
                    rrr->valid = false;
                    //  TODO tag remote as deletable
                    //  Can't remove here because need to preserve integrity
                    //    of remote[] while looping thru it
                }
                else{  //  Resend
                    //  TODO  save targetAddr in case it's different from
                    //  where this is going, and when done with this,
                    //  restore default targetAddr
                    /* CAN NOT just pass the data to hbSend()
                      because we need to retain seqno  */
                    *type  = APP_MSG;
                    *seqno = wfa_itr->seqno;
                    memcpy(package + HEADER_SIZE, &(wfa_itr->msg),wfa_itr->len);
                    takeAim(rrr->address); 
                    sendto(sockfd, package, HEADER_SIZE + wfa_itr->len, 0,
                         baseAddress, addrLen);
                    wfa_itr->timeSent = now;     
                    log && fprintf(log,
                        "    Resent msg %d\n", *seqno);
                    log && fflush(log); 
                }
            }
        }
        else{ //  Resends count as heartbeats and can ID lost connections
              //      so we only have to do these 2 jobs when wfa is empty
            //  Check for lost connection
            timeWaited = tvSubtraction(now, rrr->lastReceive);
            if(timeWaited > rrr->waitTime * 10){
                (*lostRemote)(rrr->address);
                rrr->valid = false;
                //  TODO tag remote as deletable
                //  Can't remove here because need to preserve integrity
                //    of remote[] while looping thru it
            }
            else{  
                timeWaited = tvSubtraction(now, rrr->lastSend);
                if(timeWaited > rrr->waitTime){
                    takeAim(rrr->address); 
                    *type = HEARTBEAT;
                    sendto(sockfd,package,HEADER_SIZE, 0, baseAddress, addrLen);
                    rrr->lastSend = now;
                }
            }
        }
    }
    
    return   (bytesRead)? bytesRead - HEADER_SIZE: 0;
}

//////////////////////////////////////////////////////////////////////////////
list<Remote>::iterator remoteHandle(HBaddr source, int type)
{
    list<Remote>::iterator x;
    x = find(remote.begin(), remote.end(), source);
    if(x == remote.end()){
        if(type != APP_MSG_CLESS){
            if(remote.size() < MAX_REMOTES){
                remote.push_back(source);
                x = find(remote.begin(), remote.end(), source);
                log && fprintf(log, "    Added remote\n");
            }
        }
    }

    return x;
}

//////////////////////////////////////////////////////////////////////////////
int addPremature(list<Remote>::iterator r, int sq, void* msg, int size)
{
    if(r->prm.size() == MAX_PREMATURE)
        return 0; 
    pair<set<SavedMsg>::iterator, bool> xx;
    xx = r->prm.insert(SavedMsg((char*)msg, size, sq));
    if(!xx.second)
        fputs("    Error: insert failed\n", log);
    return 1;
    // set::insert does not allow duplicates
    //     if switching to a diff container class, might have to manually
    //     check for dupes before adding msg
}

//////////////////////////////////////////////////////////////////////////////
char* HBaddr::toString(char* str)
{
//  TODO
//  THIS COULD BE MORE EFFICIENT
//  NEED TO FIND OUT HOW TO DO THIS WITHOUT CREATING THE PARENT STRUCTURE
//  inet_ntoa() WORKS ON

    struct sockaddr_in temp;
    temp.sin_addr.s_addr = host;
    sprintf(str, "%s: %d", (char*)inet_ntoa(temp.sin_addr), port);
    return str;
}

//////////////////////////////////////////////////////////////////////////////
char* hbhostToDot(hbhost ip, char* str)
{
//  TODO
//  THIS COULD BE MORE EFFICIENT
//  NEED TO FIND OUT HOW TO DO THIS WITHOUT CREATING THE PARENT STRUCTURE
//  inet_ntoa() WORKS ON
    struct sockaddr_in temp;
    temp.sin_addr.s_addr = ip;
    strcpy(str, (char*)inet_ntoa(temp.sin_addr));
    return str;
}

#ifdef BLAH
//////////////////////////////////////////////////////////////////////////////
int idDotSender(char *addr, struct sockaddr_in addrStruct)
{
//  app wouldn't use this unless it needs to print out the address
    strcpy(addr, (char*)inet_ntoa(addrStruct.sin_addr));
    return ntohs(addrStruct.sin_port);
}
# endif

//////////////////////////////////////////////////////////////////////////////
long tvSubtraction(timeval finish, timeval start)
{
    long x;
    long sec = finish.tv_sec - start.tv_sec;
    long usec;
    if(finish.tv_sec > start.tv_usec)
        usec = finish.tv_usec - start.tv_usec;
    else
        usec = 1000000 - start.tv_usec + finish.tv_usec;
    x = sec * 1000000 + usec;
    return x;
}

//////////////////////////////////////////////////////////////////////////////
void hbNetClose(void)
{
//  SHOULD TRY TO RESEND ALL LEFTOVER UNACKNOWLEDGED MESSAGES
//  OR MAYBE DOC THAT CLIENT SHOULD CHECK SEQNO BEFORE LOGOUT AND NOT
//  SHUT DOWN UNTIL LOG OUT IS ACKNOWLEDGED

    activated = false;
/*
    for(int x = 0; x < remotes; x++){
        remote[x].prm.clear();
        remote[x].wfa.clear();
    }
*/
    remote.clear();

#ifdef HBNET_WINDOWS
    closesocket(sockfd);
#else
    close(sockfd);
#endif

    log && fclose(log);
    
#ifdef HBNET_WINDOWS
    WSACleanup();
#endif
}


#ifdef HBNET_WINDOWS
void gettimeofday(struct timeval* p, void* tz /* IGNORED */){
     union {
       long long ns100; /*time since 1 Jan 1601 in 100ns units */
       FILETIME ft;
     } _now;

     GetSystemTimeAsFileTime( &(_now.ft) );
     p->tv_usec=(long)((_now.ns100 / 10LL) % 1000000LL );
     p->tv_sec= (long)((_now.ns100-(116444736000000000LL))/10000000LL);
     return;
}
#endif

